#ifndef FSTARS_H
#define FSTARS_H 1

#include "cfortran.h"
#include "Rtypes.h"

#include "Fdimpar.h"

extern "C" {
//*$ create stars.add
//*copy stars
//*
//*=== stars ============================================================*
//*
//*----------------------------------------------------------------------*
//*                                                                      *
//*     stars common for fluka:                                          *
//*                                                                      *
//*     version for fluka90/99/... of the original one of fluka86        *
//*                                                                      *
//*                                                                      *
//*     created on    15 may 1990    by    alfredo ferrari & paola sala  *
//*                                                   infn - milan       *
//*                                                                      *
//*     last change on  07-oct-99    by    alfredo ferrari               *
//*                                                                      *
//*                                                                      *
//*     description of the common variable(s):                           *
//*                                                                      *
//*        ntstar = total number of stars generated (modulo 1000000000)  *
//*        neulow = total number of low energy neutron interactions      *
//*                (modulo 1000000000)                                   *
//*        numdec = total number of decays (modulo 1000000000)           *
//*        mtstar = total number of stars generated / 100000000          *
//*        meulow = total number of low energy neutron interactions      *
//*               / 1000000000                                           *
//*        mumdec = total number of decays (modulo 1000000000)           *
//*        wstars = total weight of the stars generated                  *
//*        wneulw = total weight of the low energy neutron interactions  *
//*        westar = weight of the stars generated by                     *
//*                 different particle types                             *
//*        westop = total weight of the particles stopped                *
//*        wstop  = weight of particles of different types stopped       *
//*        weisec = total weight of the secondaries created              *
//*        weifis = total weight of high energy fissions                 *
//*        weipri = total weight of the primaries handled                *
//*        edi    = deposited energy                                     *
//*                 edi(1) = by ionisation                               *
//*                 edi(2) = by pi-zeros and/or em cascade               *
//*                 edi(3) = by nuclear excitation (or nuclear recoil    *
//*                          and heavies if the evaporation module is    *
//*                          activated)                                  *
//*                 edi(4) = by stopped particles                        *
//*                 edi(5) = energy leaving the system                   *
//*                 edi(6) = energy carried by discarded particles       *
//*                 edi(7) = by residual excitation energy (only if the  *
//*                          evaporation module is activated)            *
//*                 edi(8) = by low energy neutrons (kerma due to low    *
//*                          energy neutrons transport is in effect)     *
//*                 edi(9) = energy carried by time killed particles     *
//*                 edi(10)= energy wasted for nuclear binding energy    *
//*                          effects for e > 50 mev                      *
//*                 edi(11)= energy wasted for nuclear binding energy    *
//*                          effects for low energy neutrons             *
//*        wdec   = weight of the particles decayed                      *
//*        wdau   = weight of the decay-products                         *
//*        wtkll  = weight of the time-killed particles                  *
//*        wtdec  = total weight of the particles decayed                *
//*        wtdau  = total weight of the decay products                   *
//*        wttkll = total weight of the time killed particles            *
//*        wlwnsc = weight of the low energy neutrons interaction secon- *
//*                 daries                                               *
//*        wdecct = weight of the particles decayed for which a c tau    *
//*                 scoring is done                                      *
//*        deccts = decay c tau scoring                                  *
//*----------------------------------------------------------------------*
//*

typedef struct {
   Double_t westar[nallwp+7];
   Double_t wstop[nallwp+7];
   Double_t wdau[nallwp+7];
   Double_t wtkll[nallwp+7];
   Double_t wdec[nallwp+7];
   Double_t weifis[nallwp+7];
   Double_t deccts[nallwp+7];
   Double_t wdecct[nallwp+7];
   Double_t weisec[nallwp+19];
   Double_t wlwnsc[4];
   Double_t wophsc[4];
   Double_t edi[11];
   Double_t woptph;
   Double_t wneulw;
   Double_t westop;
   Double_t wstars;
   Double_t weipri;
   Double_t wtdec;
   Double_t wtdau;
   Double_t wttkll;
   Int_t    numdec;
   Int_t    ntstar;
   Int_t    neulow;
   Int_t    numoph;
   Int_t    mumdec;
   Int_t    mtstar;
   Int_t    meulow;
   Int_t    mumoph;
} starsCommon;
#define STARS COMMON_BLOCK(STARS,stars)
COMMON_BLOCK_DEF(starsCommon,STARS);
}

#endif
